from tensorflow.python.keras.utils import to_categorical

class Preprocessor(object):
	"""docstring for Preprocessor"""
	def __init__(self, train_samples, 
			test_samples, val_samples):
		super(Preprocessor, self).__init__()
		self.train_samples = train_samples
		self.test_samples = test_samples
		self.val_samples = val_samples

	def fit_transform(self, X, y):

		# Reshape Fetaures
		X = X.reshape(70000, 28, 28, 1)

		# Rescale Fetaures
		X = self.rescale_features(X)

		# Labels to Categorical
		y = to_categorical(y, 10)

		# Train, test, validation split
		X_train, y_train, X_test, y_test, X_val, y_val \
			= self.train_test_validation_split(X, y)

		return X_train, y_train, X_test, y_test, X_val, y_val

	def rescale_features(self, X):
		X = X.astype("float32")
		X /= 255
		return X

	def train_test_validation_split(self, features, labels):
		
		X_test = features[0:self.test_samples]
		y_test = labels[0:self.test_samples]

		X_val = features[self.test_samples:self.test_samples + self.val_samples]
		y_val = labels[self.test_samples:self.test_samples + self.val_samples]

		X_train = features[self.test_samples + self.val_samples: \
			self.test_samples + self.val_samples + self.train_samples]
		y_train = labels[self.test_samples + self.val_samples: \
			self.test_samples + self.val_samples + self.train_samples]
		
		return X_train, y_train, X_test, y_test, X_val, y_val

		



from prostagma.performances.cross_validation import CrossValidation

class SearchTechnique(object):
	"""docstring for Technique"""
	def __init__(self, parameters, model, 
			performance_validator=CrossValidation()):
		super(SearchTechnique, self).__init__()
		"""
			args:
				parameters -> dict : with the parameters to search.
				performance_validator -> Performance object : how to validate 
					the scores [default CrossValidation() with 5 fold]
				model -> function() : method to create the Keras model.
		"""
		self.parameters = parameters
		self.model = model
		self.performance_validator = performance_validator
		self.best_score = (0, 0)
		self.best_param = None

	def fit(self, X_train, y_train):
		raise NotImplementedError
		


from sklearn.utils import shuffle

from prostagma.performances.performance import Performance

class HoldOut(Performance):
	def __init__(self, val_samples=100):
		super(HoldOut, self).__init__()
		self.val_samples = val_samples
		"""
			@args
				val_samples : int -> number of samples
					used to validate the training.
		"""

	def fit(X_train, y_train, model, parameters=None):
		"""
			The method computes a simple hold out validation
			on the training data, returning the results 
			obtained.

			@args
				X_train : numpy array -> features
				y_train ; numpy array -> labels
				model : method to create the Keras model
				parameters : dictionary of parameters based on 
					the defined model
			@return
				metric : float -> score obtained on the
					validation set
		"""
		
		print("Validating the Model... ")
		X_train, y_train = shuffle(X_train, y_train)
		val_x = X_train[0:self.val_samples] 
		val_y = y_train[0:self.val_samples]
		train_x = X_train[self.val_samples:]
		train_y = y_train[self.val_samples:]
			network = model(parameters)
			network.fit(
				x=train_x, 
				y=train_y, 
				epochs=self.epochs, 
				batch_size=self.batch_size,
				verbose=0
				)
			_ , metric = network.evaluate(x=val_x, y=val_y, verbose=0)
		return metric
		


import numpy as np

from prostagma.performances.performance import Performance

class CrossValidation(Performance):
	def __init__(self, epochs=500, batch_size=32, k_fold=5):
		super(CrossValidation, self).__init__(
			epochs=epochs, batch_size=batch_size)
		"""
			@args: 
				k_fold -> int : number of fold to used.
		"""
		self.k_fold = k_fold

	def fit(self, X_train, y_train, model, parameters=None):
		"""
			The method computes k-fold cross validation
			on the training data, returning the results 
			obtained for each partition.

			@args
				X_train : numpy array -> features
				y_train ; numpy array -> labels
				model : method to create the Keras model
				parameters : dictionary of parameters based on 
					the defined model
			@return
				scores : numpy array -> all the results from 
					different partitions.
		"""

		scores = []
		val_samples = int(len(X_train) / self.k_fold)
		for fold in range(0, self.k_fold):
			print("Processing Fold: ", fold)
			val_x = X_train[fold * val_samples: (fold + 1) * val_samples] 
			val_y = y_train[fold * val_samples: (fold + 1) * val_samples]
			train_x = np.concatenate([X_train[:fold * val_samples], X_train[(fold + 1) * val_samples:]], axis=0)
			train_y = np.concatenate([y_train[:fold * val_samples], y_train[(fold + 1) * val_samples:]], axis=0)
			network = model(parameters)
			network.fit(
				x=train_x, 
				y=train_y, 
				epochs=self.epochs, 
				batch_size=self.batch_size,
				verbose=0
				)
			_ , metric = network.evaluate(x=val_x, y=val_y, verbose=0)
			scores.append(metric)
		scores = np.asarray(scores)
		return scores






class Performance(object):
	"""docstring for Perfomance"""
	def __init__(self, epochs, batch_size):
		super(Performance, self).__init__()
		self.epochs = epochs
		self.batch_size = batch_size

	def fit(self, X_train, y_train, model, parameters=None):
		raise NotImplementedError
		


import numpy as np

from keras.datasets import mnist

class Loader(object):
	"""docstring for Loader"""
	def __init__(self, number_of_samples):
		super(Loader, self).__init__()
		self.number_of_samples = number_of_samples

	def load_data(self):
		(X_train, y_train), (X_val, y_val) = mnist.load_data()
		X = np.concatenate((X_train, X_val), axis=0)
		y = np.concatenate((y_train, y_val), axis=0)
		return X, y
		


import numpy as np

from itertools import product

from prostagma.techniques.technique import SearchTechnique
from prostagma.performances.cross_validation import CrossValidation
from sklearn.grid_search import ParameterGrid

class GridSearch(SearchTechnique):
	"""
		The class implement the simple Grid Search 
		algorithm to find the best parameters using 
		Cross Validation. 
	"""
	def __init__(self, parameters, model, 
			performance_validator=CrossValidation()):
		super(GridSearch, self).__init__(parameters=parameters, 
			model=model, performance_validator=performance_validator)

	def fit(self, X_train, y_train):
		"""
			The method computes a score for each combination of 
			hyperparameters using the performance validator

			@args
				X_train : numpy array -> features
				y_train : numpy array -> labels

			@return
				all_scores : numpy array -> all the mean scores
					obtained using the performance validator	
		"""
		all_scores = []
		grid = ParameterGrid(self.parameters)
		for params in grid:
			print("Validating the model with: ", params)
			scores = self.performance_validator.fit(X_train, y_train, self.model, params)
			all_scores.append((scores.mean(), scores.std()))

			if scores.mean() > self.best_score[0]:
				self.best_score = (scores.mean(), scores.std())
				self.best_param = params
		all_scores = np.asarray(scores)
		return all_scores



		


from functools import reduce
from operator import add
import random

from prostagma.techniques.technique import SearchTechnique
from prostagma.performances.cross_validation import CrossValidation

class EvolutionaryStrategy(SearchTechnique):
	def __init__(self, parameters, model, 
			performance_validator=CrossValidation(), 
			retain=0.3, random_select=0.1, mutate_prob=0.2, 
			generations=10, population_size=20):
		super(EvolutionaryStrategy, self).__init__(parameters=parameters, 
			model=model, performance_validator=performance_validator)
		self.retain = retain
		self.random_select = random_select
		self.mutate_prob = mutate_prob
		self.generations = generations
		self.population_size = population_size

	def fit(self, X_train, y_train):
		"""
			The method calls the performance validator defined
			and uses it to compute the score for each set of parameters.
			
			@args
				X_train : ndarray -> features
				y_train : ndarray -> labels
			
			@return
				scores : dict/list with all the scores
					for each set of parameters
		"""

		# Create the Population
		population = self.create_population()

		# Train the Genetic Algorithm
		for i in range(self.generations):
			print("Generation {}" .format(i))

			for model in population:
				means, stdv = performance_validator.fit(X_train, y_train, model)

			if i < self.generations-1:
				s = self.evolve(population)
		return

	def create_population(self):
		"""
			The method creates the different networks (which are 
			the different models with different parameters).
			@args
				count : int -> number of elements that you want 
					in the population
			@return
				population list[KerasModel] 

		"""
		population = []
		for _ in range(0, self.population_size):
			parameters = self.set_random_parameters()
			population.append(self.model(parameters=parameters))
		return population

	def get_fitness(self, model):
		return model.accuracy

	def get_grade(self, population):
		""" The method returns the avg accuracy in the population 

			@args
				population : list[Model] 
			
			@return
				avg_score : float -> avg score of the population
				
		"""
		total = reduce(add, (self.get_fitness(network) for network in population))
		avg_score = float(total) / len(population)
		return avg_score

	def breed(self, mother, father):
		"""
			The method computes the crossover between two
			population.

			@args
				mother : list[Model]
				father : list[Model]
			
			@return
				children : list[Model]
		"""
		children = []
		for _ in range(2):
			child = {}
			for param in self.parameters:
				child[param] = random.choice([mother.network[param], father.network[param]])
			model = self.model()
			if self.mutate_chance > random.random():
				model = self.mutate(model)
			children.append(model)
		return children

	def mutate(self, model):
		mutation = random.choice(list(self.parameters.keys()))
		model.model[mutation] = random.choice(self.parameters[mutation])
		return model

	def evolve(self, pop):
		graded = [(self.get_fitness(model), model) for network in pop]
		graded = [x[1] for x in sorted(graded, key=lambda x: x[0], reverse=True)]
		retain_length = int(len(graded)*self.retain)

		parents = graded[:retain_length]

		for individual in graded[retain_length:]:
			if self.random_select > random.random():
				parents.append(individual)

		parents_length = len(parents)
		desired_length = len(pop) - parents_length
		children = []

		while len(children) < desired_length:
			male = random.randint(0, parents_length-1)
			female = random.randint(0, parents_length-1)

			if male != female:
				male = parents[male]
				female = parents[female]

				children_new = self.breed(male, female)

				for child_new in children_new:
					if len(children) < desired_length:
						children.append(child_new)

		parents.extend(children)
		return parents

	def get_population_accuracy(self, population):
		total_accuracy = 0
		for model in population:
			total_accuracy += network.get_accuracy

		avg_accuracy = total_accuracy / len(population)
		return avg_accuracy

	def set_random_parameters(self):
		"""
			The method should create a sample of parameters,
			given the parameters space

			@return
				parameters : dict -> dictionary with the parameters 
					to create a network. 
		"""
		raise NotImplementedError


from distutils.core import setup
from setuptools import setup, find_packages

def check_dependencies():
    install_requires = []
    
    try:
        import numpy
    except ImportError:
        install_requires.append('numpy')
    try:
        import scipy
    except ImportError:
        install_requires.append('scipy')
    try:
        import sklearn
    except ImportError:
        install_requires.append('scikit-learn')
    try:
        import pandas
    except ImportError:
        install_requires.append('pandas')
    try:
        import tensorflow
    except ImportError:
        install_requires.append('tensorflow')
    try:
        import keras
    except ImportError:
        install_requires.append('keras')
    return install_requires

install_requires = check_dependencies()

setup(
  name = 'prostagma',
  version = '0.3',
  description = 'Hyperparameters Tuning Library',
  author = 'Skopos-team',
  packages=find_packages(),
  install_requires=install_requires,
  author_email = 'skopos.library@gmail.com',
  url = 'https://github.com/Skopos-team/Prostagma', 
  license='Apache2',
  download_url = 'https://github.com/Skopos-team/Prostagma/archive/0.3.tar.gz',
  keywords = ['testing', 'logging', 'example'],
  classifiers = ['Programming Language :: Python :: 3.5',
                  'Operating System :: POSIX',
                  'Operating System :: Unix',
                  'Operating System :: MacOS'
                  ],
)


from tensorflow.python.keras.models import Sequential
from tensorflow.python.keras.layers import Dense, Dropout, Conv2D, Activation, MaxPooling2D, Flatten
from tensorflow.python.keras.optimizers import Adam
from tensorflow.python.keras.callbacks import ModelCheckpoint, TensorBoard

class Model(object):
    def __init__(self, log_dir=None, model_path=None):
        super(Model, self).__init__()
        """
			self.accuracy represents how good is a certain model. 
			When trained/tested the variable is updated based
			on the validation metric.
        """
        self.log_dir = log_dir
        self.model_path = model_path
        self.model = self.build_model()
        self.checkpointers = self.build_checkpointers()

    def build_model(self, parameters=None):        
        
        # Get parameters
        try:
            dropout = parameters["dropout"]
            learning_rate = parameters["learning_rate"]
        except:
            dropout = 0.1
            learning_rate = 0.001

        # Define Model architecture
        model = Sequential()
        model.add(Conv2D(32, (3, 3), padding="SAME", input_shape=(28, 28, 1)))
        model.add(Activation("relu"))
        model.add(Conv2D(32, (3, 3)))
        model.add(Activation("relu"))
        model.add(MaxPooling2D(pool_size=(2, 2)))
        model.add(Dropout(dropout))

        model.add(Conv2D(64, (3, 3), padding="SAME"))
        model.add(Activation("relu"))
        model.add(Conv2D(64, (3, 3)))
        model.add(Activation("relu"))
        model.add(MaxPooling2D(pool_size=(2, 2)))
        model.add(Dropout(dropout))

        model.add(Flatten())
        model.add(Dense(512))
        model.add(Activation("relu"))
        model.add(Dropout(dropout))
        model.add(Dense(10))
        model.add(Activation("softmax"))

        # Compile the model
        model.compile(
            optimizer=Adam(learning_rate),
            loss="categorical_crossentropy", 
            metrics=["accuracy"])

        return model

    def build_checkpointers(self):
        checkpointers = []
        if self.model_path is not None:
            checkpointer = ModelCheckpoint(filepath=self.model_path + '.h5', 
                monitor='val_loss', verbose=1, save_best_only=True, period=10)
            checkpointers.append(checkpointer)
        
        if self.log_dir is not None:
            tensorboard = TensorBoard(log_dir=self.log_dir, histogram_freq=0, 
                batch_size=32, write_graph=True, write_grads=False, write_images=False)
            checkpointers.append(tensorboard)

        return checkpointers

    def fit(self, X_train, y_train, epochs, batch_size, X_val=None, y_val=None):
        history = self.model.fit(
            X_train, y_train,
            epochs=epochs,
            validation_data=(X_val, y_val),
            batch_size=batch_size,
            verbose=1,
            callbacks=self.checkpointers)

        return history

    def evaluate(self, X_test, y_test):
        loss, metric = self.model.evaluate(X_test, y_test, verbose=0)
        return loss, metric

    def predict(self, X):
        y_pred = self.model.predict(X)
        return y_pred

    def restore_weights(self):
        try:
            self.model.load_weights(self.model_path + ".h5")
            print("Loaded model from disk")
        except OSError:
            pass
        return

    def run_model(self, X_train, y_train, X_test, y_test, 
        X_val, y_val, train, test):
        
        # Train the model
        if train == True:
            history = self.fit(X_train, y_train, X_val, y_val)
        
        # Test the model
        if test == True:
            score = self.evaluate(X_test, y_test)
            print("SCORE: ", score)


        return
		